/* acre - A cairo-based library for creating plots and charts.
 *
 * Copyright © 2009 Carl Worth <cworth@cworth.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include "acre.h"
#include "xmalloc.h"

#include <string.h>
#include <math.h>

typedef struct _acre_data_point_2d {
    double x;
    double y;
} acre_data_point_2d_t;

struct _acre_data {
    char *name;

    acre_data_point_2d_t *points;
    unsigned int points_size;
    unsigned int num_points;
};

struct _acre {
    char *title;
    char *x_axis_label;
    char *y_axis_label;

    acre_data_t **data;
    unsigned int data_size;
    unsigned int num_data;
};

/* Create a new, empty plot. */
acre_t *
acre_create (void)
{
    acre_t *acre;

    acre = xmalloc (sizeof (acre_t));

    acre->title = NULL;
    acre->x_axis_label = NULL;
    acre->y_axis_label = NULL;

    acre->data = NULL;
    acre->data_size = 0;
    acre->num_data = 0;

    return acre;
}

/* Destroy a plot. */
void
acre_destroy (acre_t *acre)
{
    unsigned int i;

    free (acre->title);
    free (acre->x_axis_label);
    free (acre->y_axis_label);

    for (i = 0; i < acre->num_data; i++)
	acre_data_destroy (acre->data[i]);

    free (acre->data);

    free (acre);
}

void
acre_set_title (acre_t *acre, const char *title)
{
    free (acre->title);

    acre->title = strdup (title);
}

void
acre_set_x_axis_label (acre_t *acre, const char *label)
{
    free (acre->x_axis_label);

    acre->x_axis_label = strdup (label);
}

void
acre_set_y_axis_label (acre_t *acre, const char *label)
{
    free (acre->y_axis_label);

    acre->y_axis_label = strdup (label);
}

/* Add a dataset to the plot. The plot assumes ownership of the
 * dataset so it is not necessary to call acre_data_destroy on it. */
void
acre_add_data (acre_t *acre, acre_data_t *data)
{
    if (acre->num_data >= acre->data_size) {
	acre->data_size *= 2;
	if (acre->data_size == 0)
	    acre->data_size = 1;
	acre->data = xrealloc_ab (acre->data,
				  acre->data_size,
				  sizeof (acre_data_t *));
    }

    acre->data[acre->num_data] = data;
    acre->num_data++;
}

#define ACRE_FONT_FAMILY "sans"
#define ACRE_FONT_SIZE 12
#define ACRE_TITLE_FONT_SIZE 32
#define ACRE_PAD (ACRE_FONT_SIZE)
#define ACRE_TICK_SIZE 4

/* Draw the plot to the given cairo context within a user-space
 * rectangle from (0, 0) to (width, height). This size includes all
 * space for extra-plot elements (such as the title, the axis labels,
 * etc.)
 */
void
acre_draw (acre_t *acre, cairo_t *cr, double width, double height)
{
    PangoFontDescription *acre_font, *title_font;
    PangoLayout *title_layout, *x_axis_layout, *y_axis_layout;
    int title_width, title_height;
    int x_axis_width, x_axis_height;
    int y_axis_width, y_axis_height;
    PangoRectangle chart, new_chart;

    cairo_save (cr);

    cairo_set_source_rgb (cr, 1, 1, 1);

    cairo_paint (cr);

    acre_font = pango_font_description_new ();
    pango_font_description_set_family (acre_font, ACRE_FONT_FAMILY);
    pango_font_description_set_absolute_size (acre_font,
					      ACRE_FONT_SIZE * PANGO_SCALE);

    title_font = pango_font_description_new ();
    pango_font_description_set_family (title_font, ACRE_FONT_FAMILY);
    pango_font_description_set_absolute_size (title_font,
					      ACRE_TITLE_FONT_SIZE * PANGO_SCALE);

    title_layout = pango_cairo_create_layout (cr);
    pango_layout_set_font_description (title_layout, title_font);
    pango_layout_set_text (title_layout, acre->title, -1);
    pango_layout_set_alignment (title_layout, PANGO_ALIGN_CENTER);

    x_axis_layout = pango_cairo_create_layout (cr);
    pango_layout_set_font_description (x_axis_layout, acre_font);
    pango_layout_set_text (x_axis_layout, acre->x_axis_label, -1);
    pango_layout_set_alignment (x_axis_layout, PANGO_ALIGN_CENTER);

    y_axis_layout = pango_cairo_create_layout (cr);
    pango_layout_set_font_description (y_axis_layout, acre_font);
    pango_layout_set_text (y_axis_layout, acre->y_axis_label, -1);
    pango_layout_set_alignment (y_axis_layout, PANGO_ALIGN_CENTER);

    /* Iterate with the layout of the title and axis labels until they
     * are stable, (this requires iteration since we don't know what
     * to set their widths to in advance due to the wrapping of the
     * other elements). */
    chart.x = 0;
    chart.y = 0;
    chart.width = width;
    chart.height = height;
    while (1) {
	pango_layout_set_width (title_layout, chart.width * PANGO_SCALE);
	pango_layout_set_width (x_axis_layout, chart.width * PANGO_SCALE);
	pango_layout_set_width (y_axis_layout, chart.height * PANGO_SCALE);

	pango_layout_get_pixel_size (title_layout, &title_width, &title_height);
	pango_layout_get_pixel_size (x_axis_layout, &x_axis_width, &x_axis_height);
	pango_layout_get_pixel_size (y_axis_layout, &y_axis_width, &y_axis_height);

	new_chart.x = ACRE_PAD + y_axis_height +
	    ACRE_PAD + ACRE_FONT_SIZE;
	new_chart.width = width - chart.x - ACRE_PAD;

	new_chart.y = ACRE_PAD + title_height + ACRE_PAD;
	new_chart.height = height - chart.y - (ACRE_FONT_SIZE + ACRE_PAD + x_axis_height + ACRE_PAD);

	if (new_chart.x == chart.x &&
	    new_chart.y == chart.y &&
	    new_chart.width == chart.width &&
	    new_chart.height == chart.height)
	{
	    break;
	}

	chart.x = new_chart.x;
	chart.y = new_chart.y;
	chart.width = new_chart.width;
	chart.height = new_chart.height;
    }

    cairo_set_source_rgb (cr, 0, 0, 0);

    cairo_move_to (cr, chart.x, ACRE_PAD);
    pango_cairo_show_layout (cr, title_layout);

    cairo_save (cr);
    {
	cairo_translate (cr, ACRE_PAD, chart.y + chart.height);
	cairo_rotate (cr, - M_PI / 2.0);
	cairo_move_to (cr, 0, 0);
	pango_cairo_show_layout (cr, y_axis_layout);
    }
    cairo_restore (cr);

    cairo_move_to (cr, chart.x,
		   chart.y + chart.height + ACRE_FONT_SIZE + ACRE_PAD);
    pango_cairo_show_layout (cr, x_axis_layout);

    cairo_rectangle (cr,
		     chart.x - 0.5, chart.y - 0.5,
		     chart.width + 1.0, chart.height + 1.0);
    cairo_set_line_width (cr, 1.0);
    cairo_stroke (cr);
}

/* Create a new dataset---a collection of (x, y) datapoints. A single
 * plot can contain multiple datasets, (see acre_add_data). */
acre_data_t *
acre_data_create (void)
{
    acre_data_t *data;

    data = xmalloc (sizeof (acre_data_t));

    data->name = NULL;

    data->points = NULL;
    data->points_size = 0;
    data->num_points = 0;

    return data;
}

/* Destroy an acre dataset. Do not call this function if the dataset
 * has been added to an acre_t plot with acre_add_data. */
void
acre_data_destroy (acre_data_t *data)
{
    free (data->points);

    free (data);
}

/* Set the label for this dataset (to appear in the plot's key). */
void
acre_data_set_name (acre_data_t *data, const char *name)
{
    free (data->name);

    data->name = strdup (name);
}

/* Add a datapoint to the given dataset. */
void
acre_data_add_point_2d (acre_data_t *data, double x, double y)
{
    if (data->num_points >= data->points_size) {
	data->points_size *= 2;
	if (data->points_size == 0)
	    data->points_size = 16;
	data->points = xrealloc_ab (data->points,
				    data->points_size,
				    sizeof (acre_data_point_2d_t));
    }

    data->points[data->num_points].x = x;
    data->points[data->num_points].y = y;
    data->num_points++;
}
